/* 
 * $COPYRIGHT$
 * $Id: MessageFormatter.java,v 1.4 2001/09/23 23:49:53 neuro Exp $
 *
 * Date        Author            Changes 
 * May 08 2001 Remus Pereni      Created
 */
package org.media.mn8;


import java.util.Properties;
import org.media.mn8.event.*;
/**
 * Class responsible for formatting logging messages. Loading them from the resource 
 * bundles and properly formatting them to be displayed.
 * @author <a href="mailto:remus@nolimits.ro">Remus Pereni</a>
 * @version $Revision: 1.4 $ $Date: 2001/09/23 23:49:53 $
 */
public class MessageFormatter {

    private static final String _MESSAGES_RESOURCE_FILE = "conf/res/mn8messages.properties";

    private static Properties _prop;

    private static String _head = "mn8 : ";
    private static String _warning = "[WARNING] - ";
    private static String _error = "[ERROR] - ";
    private static String _info = "[INFO] - ";


    private static MessageFormatter _formatter = new MessageFormatter();
    
    /**
     * Dafault constructor
     */
    private MessageFormatter () {        
        try {
            _prop = new Properties();
            Class c = Class.forName("org.media.mn8.LogManager");

            if ( c.getResourceAsStream( _MESSAGES_RESOURCE_FILE ) == null)
                throw new RuntimeException ("Unable to load resource file : " 
                                            + _MESSAGES_RESOURCE_FILE);

            _prop.load(c.getResourceAsStream( _MESSAGES_RESOURCE_FILE ));           

            if ( _prop.getProperty ("mn8.messages.head") != null) 
                _head = _prop.getProperty ("mn8.messages.head") ;

            if ( _prop.getProperty ("mn8.messages.type.warning") != null) 
                _warning = _prop.getProperty ("mn8.messages.type.warning") ;

            if ( _prop.getProperty ("mn8.messages.type.error") != null) 
                _error = _prop.getProperty ("mn8.messages.type.error") ;

        } catch (Exception ex) {
            sysError ( _warning, ex.getMessage());
            _prop =  null;
        }
    }


    /**
     * Method for getting the MessageFormatter instance. Don't forger that it
     * implements a Singleton pattern.
     * @return The default instance of MessageFormatter
     */
    public static MessageFormatter getReference () {
        return _formatter;
    }


    /**
     * The main method for displaying system messages for mn8.
     * @param type_id The type of the messages (error, warning, ...). This is
     * actually an key in the resource bundle where the actual value which will be
     * displayed is hold. mn8.messages.type. has to be inserted before the key in the resource 
     * bundle to store such messages types.
     * <pre>
     *   mn8.messages.type.warning=[WARNING]
     * </pre>
     * @param message_id The id of the messages. This is
     * actually an key in the resource bundle where the actual value which will be
     * displayed is hold. mn8.messages. has to be inserted before the key in the resource 
     * bundle to store such messages.
     * <pre>
     *   mn8.messages.file_missing=Sorry the file: xxx.jar \n is mising.
     * </pre>
     */
    public String getMessage(String type_id, String message_id ) {
        return getMessage(type_id, message_id, "", "");
    }


    /**
     * The main method for displaying system messages for mn8.
     * @param type_id The type of the messages (error, warning, ...). This is
     * actually an key in the resource bundle where the actual value which will be
     * displayed is hold. mn8.messages.type. has to be inserted before the key in the resource 
     * bundle to store such messages types.
     * <pre>
     *   mn8.messages.type.warning=[WARNING]
     * </pre>
     * @param message_id The id of the messages. This is
     * actually an key in the resource bundle where the actual value which will be
     * displayed is hold. mn8.messages. has to be inserted before the key in the resource 
     * bundle to store such messages.
     * <pre>
     *   mn8.messages.file_missing=Sorry the file: xxx.jar \n is mising.
     * </pre>
     * @param optional Another text to be added to the message. Usualy displayed on the same line.
     * @return The formatted message.
     */
    public String getMessage(String type_id, String message_id, String optional) {
        return getMessage( type_id, message_id, optional, "");
    }


    /**
     * The main method for displaying system interpretation related messages for mn8.
     * @param type_id The type of the messages (error, warning, ...). This is
     * actually an key in the resource bundle where the actual value which will be
     * displayed is hold. mn8.messages.type. has to be inserted before the key in the resource 
     * bundle to store such messages types.
     * <pre>
     *   mn8.messages.type.warning=[WARNING]
     * </pre>
     * @param message_id The id of the messages. This is
     * actually an key in the resource bundle where the actual value which will be
     * displayed is hold. mn8.messages. has to be inserted before the key in the resource 
     * bundle to store such messages.
     * <pre>
     *   mn8.messages.file_missing=Sorry the file: xxx.jar \n is mising.
     * </pre>
     * @param optional Another text to be added to the message. Usualy displayed on the same line.
     * @param line The line number of the erroneous message.
     * @return The formatted message.
     */
    public String getInterpMessage(String type_id, String message_id, String optional, int line) {
        if( optional == null ) optional = "";

        return _head 
            + getTypeInfo( type_id ) 
            + "Line: " + line + " "
            + getMessageInfo( message_id ) 
            + optional;
    }


    /**
     * The main method for displaying system messages for mn8.
     * @param type_id The type of the messages (error, warning, ...). This is
     * actually an key in the resource bundle where the actual value which will be
     * displayed is hold. mn8.messages.type. has to be inserted before the key in the resource 
     * bundle to store such messages types.
     * <pre>
     *   mn8.messages.type.warning=[WARNING]
     * </pre>
     * @param message_id The id of the messages. This is
     * actually an key in the resource bundle where the actual value which will be
     * displayed is hold. mn8.messages. has to be inserted before the key in the resource 
     * bundle to store such messages.
     * <pre>
     *   mn8.messages.file_missing=Sorry the file: xxx.jar \n is mising.
     * </pre>
     * @param optional Another text to be added to the message. Usualy displayed on the same line.
     * @param suplement Another text to be added to the message. Usualy displayed on another line.
     * @return The formatted message.
     */
    public String getMessage(String type_id, String message_id, String optional, String suplementary ) {
        if( optional == null ) optional = "";
        if( suplementary == null ) suplementary = "";

        return _head 
            + getTypeInfo( type_id ) 
            + getMessageInfo( message_id ) 
            + optional + "\n" + suplementary;
    }


    public String getMessageInfo( String message_id ) {
        if ( message_id == null || message_id.equals("") ) return "";

        if ( _prop.getProperty("mn8.messages." + message_id) == null ){
            sysError (_warning, "Unable to find mn8.messages." + message_id 
                      + " in resource bundle: " + _MESSAGES_RESOURCE_FILE);

            return "";
        } else return _prop.getProperty("mn8.messages." + message_id);
        
    }


    public String getTypeInfo( String type_id ) {
        if( type_id == null || type_id.equals("") ) return "";

        if ( _prop.getProperty("mn8.messages.type." + type_id) == null ){
            sysError (_warning, "Unable to find mn8.messages.type." + type_id
                      + " in resource bundle: " + _MESSAGES_RESOURCE_FILE);
            return "";
        } else return _prop.getProperty("mn8.messages.type." + type_id);
        
    }

    
    private void sysError (String type, String message ) {
        throw new mn8SystemException( _head + type + message );
    }


    private void sysError (String type, String message, String optional ) {
        throw new mn8SystemException( _head + type + message + optional );
    }


}
