/* 
 * $COPYRIGHT$
 * $Id: NamingManager.java,v 1.10 2002/03/25 20:01:36 neuro Exp $
 *
 * Date        Author            Changes 
 * May 18 2001 Remus Pereni      Created
 */
package org.media.mn8;

import org.media.mn8.conf.NamingConf;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingEnumeration;
import javax.naming.NamingException;
import javax.naming.NameNotFoundException;
import javax.naming.OperationNotSupportedException;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.StringTokenizer;

/**
 * Class who is responsible for wrapping and managing the 
 * JNDI Context specified in conf/naming.mn8.xml.
 * @author <a href="mailto:remus@nolimits.ro">Remus Pereni</a>
 * @version $Revision: 1.10 $ $Date: 2002/03/25 20:01:36 $
 */
public class NamingManager {

    public static final String ENV_ID = "env:";
    public static final String MEM_ID = "mem:";

    private static final String _ENV_CONTEXT_PATH = "env";
    private static final String _MEM_CONTEXT_PATH = "mem";   

    private static NamingManager _nManagerInstance = new NamingManager();
    private static Context _initialContext ;
    private static Context _envContext;
    private static Context _memContext;
    private static String _memMapping = "/";
    private static String _envMapping = "/";


    private NamingManager () {
        try {
            NamingConf _config = NamingConf.getReference();
            _memMapping = _config.getMemMapping();
            _envMapping = _config.getEnvMapping();     

            _initialContext = new InitialContext ( _config.getInitialContextProperties() );
            _memMapping += _MEM_CONTEXT_PATH;
            _envMapping += _ENV_CONTEXT_PATH;

            _envContext = testAndCreateContext( _initialContext,
                                                _envMapping);
            _memContext = testAndCreateContext( _initialContext,
                                                _memMapping);
            initializeEnvContext ( _config.getEnvProperties() );
        } catch (NamingException nex) {            
            LogManager.message( "error",
                                "naming.create_exception",
                                nex.getMessage());
            System.exit(0);
        }
    }

  
    /**
     * Method for getting the NamingManager instance. Don't forget that it
     * implements a Singleton pattern.
     * @return The default instance of NamingManager
     */
    public static NamingManager getReference () {
        return _nManagerInstance;
    }


    /**
     * Binds a name to an object.  The scheme must be specified 
     * ("mem:" or "env:")!
     * @param name The name to bind
     * @param obj The object to be binded.
     * @throws javax.naming.NamingException If a naming exception is 
     * encountered.
     */
    public static void bind (String name, Object obj)
        throws NamingException {

        String bndName = name;
        Context context = _memContext;

        if ( name == null ||
             obj == null ) 
            throw new RuntimeException();

        if ( name.startsWith( ENV_ID ) )
            throw new OperationNotSupportedException( "Context is read-only" );
        else 
            if ( name.startsWith( MEM_ID ) ) {
                name = name.substring(MEM_ID.length());
                
                if ( name.lastIndexOf("/") != -1 &&
                     name.substring(0, name.lastIndexOf("/")).length() != 0 ) {
                    bndName = name.substring(name.lastIndexOf("/") + 1);
                    context = testAndCreateContext(_memContext,  
                                                   name.substring(0, name.lastIndexOf("/")));
                }
                bndName = bndName.substring(bndName.lastIndexOf("/") + 1);
                context.bind(bndName, obj);

            } 
            else 
                throw new NameNotFoundException();            
    }


    /**
     * Rebinds or binds a name to an object.  The scheme must be specified 
     * ("mem:" or "env:")!
     * @param name The name to bind
     * @param obj The object to be binded.
     * @throws javax.naming.NamingException If a naming exception is 
     * encountered.
     */
    public static void rebind (String name, Object obj)
        throws NamingException {

        String bndName = name;
        Context context = _memContext;

        if ( name == null ||
             obj == null ) 
            throw new RuntimeException();

        if ( name.startsWith( ENV_ID ) )
            throw new OperationNotSupportedException( "Context is read-only" );
        else 
            if ( name.startsWith( MEM_ID ) ) {
                name = name.substring(MEM_ID.length());
                
                if ( name.lastIndexOf("/") != -1 &&
                     name.substring(0, name.lastIndexOf("/")).length() != 0 ) {
                    bndName = name.substring(name.lastIndexOf("/") + 1);
                    context = testAndCreateContext(_memContext,  
                                                   name.substring(0, name.lastIndexOf("/")));
                }
                bndName = bndName.substring(bndName.lastIndexOf("/") + 1);
                context.rebind(bndName, obj);

            } 
            else 
                throw new NameNotFoundException();            
    }


    /**
     * Binds a new name to the object bound to an old name, and unbinds the old
     * name. The scheme must be specified ("mem:" or "env:")! Env: is read-only.
     * @param oldName The name of the existing binding. May not be empty.
     * @param newName The name of the new binding. May not be empty.
     * @throws javax.naming.NamingException If a naming exception is 
     * encountered.
     */
    public static void rename( String oldName, String newName )
        throws NamingException {

        Object obj;

        if ( oldName == null ||
             newName == null ||
             oldName == ""   ||
             newName == "" ) 
            throw new RuntimeException();

        if ( oldName.startsWith( ENV_ID ) ||
             newName.startsWith( ENV_ID ) )
            throw new OperationNotSupportedException( "Context is read-only" );
        else 
            if ( oldName.startsWith( MEM_ID ) && 
                 newName.startsWith( MEM_ID ) ) {

                obj = lookup( oldName );
                bind( newName, obj );
                unbind( oldName );
            } 
            else 
                throw new NameNotFoundException();            
    }


    /**
     * Unbinds the object binded under the name.
     * The scheme must be specified ("mem:" or "env:")! Env: is read-only.
     * @param name The name to unbind. May not be empty.
     * @throws javax.naming.NamingException If a naming exception is 
     * encountered.
     */
    public static void unbind( String name )
        throws NamingException {

        Object obj;

        if ( name == null ||
             name == ""  )
            throw new RuntimeException();

        if ( name.startsWith( ENV_ID ) )
            throw new OperationNotSupportedException( "Context is read-only" );
        else 
            if ( name.startsWith( MEM_ID ) ) {
                name = name.substring(MEM_ID.length());
                _memContext.unbind( name );
            } 
            else 
                throw new NameNotFoundException();            
    }


    /**
     * Enumeratest the names bound in the named context.
     * The scheme must be specified ("mem:" or "env:")!
     * @param name The name of the context to list.
     * @return An enumeration of the names and class names
     * of the bindings in this context.
     * @throws javax.naming.NamingException If a naming exception is 
     * encountered.
     */
    public static NamingEnumeration list( String name )
        throws NamingException {

        if ( name == null ||
             name == ""  )
            throw new RuntimeException();

        if ( name.startsWith( MEM_ID ) ) {
            name = name.substring(MEM_ID.length());
            return  _memContext.list( name );
        } else 
            if ( name.startsWith( ENV_ID ) ) { 
                name = name.substring( ENV_ID.length() );
                return  _envContext.list( name );
            }
            else throw new NameNotFoundException();                 
    }


    /**
     * Enumeratest the names bound in the named context, along with the
     * objects bound to them. The contens of any subcontext is not included.
     * The scheme must be specified ("mem:" or "env:")!
     * @param name The name of the context to list.
     * @return An enumeration of the bindings in this context.
     * Each element of the enumeration is of type Binding.
     * @throws javax.naming.NamingException If a naming exception is 
     * encountered.
     * @see javax.naming.Binding
     */
    public static NamingEnumeration listBindings( String name )
        throws NamingException {

        if ( name == null ||
             name == ""  )
            throw new RuntimeException();

        if ( name.startsWith( MEM_ID ) ) {
            name = name.substring(MEM_ID.length());
            return  _memContext.listBindings( name );
        } else 
            if ( name.startsWith( ENV_ID ) ) { 
                name = name.substring( ENV_ID.length() );
                return  _envContext.listBindings( name );
            }
            else throw new NameNotFoundException();                 
    }


    /**
     * Creates and binds a new context.
     * The scheme must be specified 
     * ("mem:" or "env:")! However the env: schema is read-only.
     * @param name The name of the subcontext.
     * @throws javax.naming.NamingException If a naming exception is 
     * encountered.
     */
    public static Context createSubcontext( String name ) 
        throws NamingException{

        Context context;

        if ( name == null )
            throw new RuntimeException();

        if ( name.startsWith( ENV_ID ) )
            throw new OperationNotSupportedException( "Context is read-only" );
        else 
            if ( name.startsWith( MEM_ID ) ) {
                name = name.substring(MEM_ID.length());                
                context = testAndCreateContext(_memContext, name);
                return context;
            } 
            else 
                throw new NameNotFoundException();            
    }


    /**
     * Destroys the named  context.
     * The scheme must be specified 
     * ("mem:" or "env:")! However the env: schema is read-only.
     * @param name The name of the subcontext.
     * @throws javax.naming.NamingException If a naming exception is 
     * encountered.
     */
    public static void destroySubcontext( String name ) 
        throws NamingException{

        Context context;

        if ( name == null )
            throw new RuntimeException();

        if ( name.startsWith( ENV_ID ) )
            throw new OperationNotSupportedException( "Context is read-only" );
        else 
            if ( name.startsWith( MEM_ID ) ) {
                name = name.substring(MEM_ID.length());
                _memContext.destroySubcontext( name );
            } 
            else 
                throw new NameNotFoundException();            
    }


    /**
     * Closes this context.
     * @throws javax.naming.NamingException If a naming exception is 
     * encountered.
     */
    public static void close() throws NamingException {
        _memContext.close();
        _envContext.close();
        _memContext = null;
        _envContext = null;
    }


    /**
     * Retrieves the named object. The scheme must be specified 
     * ("mem:" or "env:")!
     * @param name The name of the object to be looked up.
     * @throws javax.naming.NamingException If a naming exception is 
     * encountered.
     */
    public static Object lookup (String name) 
        throws NamingException {

        if ( name.startsWith( ENV_ID ) ) {
            name = name.substring(ENV_ID.length());
            return _envContext.lookup(name);
        } 
        else 
            if ( name.startsWith( MEM_ID ) ) {
                name = name.substring(MEM_ID.length());
                return _memContext.lookup(name);
            } 
            else 
                return new NameNotFoundException();            
    }


    /**
     * Because from the outside the env:/ context is read only
     * this is the default method for populating it. <br>
     * The name can be specified either as: "env:/somecontext/../name"
     * or without the "env:" specifier as: "somecontext/../name".<br>
     * It will properly create the subcontextes if there doesn't exist.
     * @param name The name under which the object will be binded in
     * the directory context.
     * @param obj The object which will be binded.
     * @throws javax.naming.NamingException If a naming exception is 
     * encountered.
     */
    public static void envBind(String name, Object obj)
        throws NamingException {
        
        String bndName = name;
        Context context = _envContext;
        if ( name == null ||
             obj == null ) 
            throw new RuntimeException();
        
        if ( name.startsWith( ENV_ID ) )
            name = name.substring(ENV_ID.length());

        if ( name.startsWith( MEM_ID ) )
            throw new OperationNotSupportedException( MEM_ID + " unsupported scheme!" );
        
        if ( name.lastIndexOf("/") != -1 &&
             name.substring(0, name.lastIndexOf("/")).length() != 0 ) {
            bndName = name.substring(name.lastIndexOf("/") + 1);
            context = testAndCreateContext(_envContext,  
                                           name.substring(0, name.lastIndexOf("/")));
        }
        bndName = bndName.substring(bndName.lastIndexOf("/") + 1);
        context.rebind(bndName, obj);        

    }


    /**
     * Parses the naming path (ex: "/system/vars/memmory") and if there are no
     * correspondent contextes, creates them.
     */
    private static Context testAndCreateContext(Context startCtx, 
                                                String subcontextPath ) 
        throws NamingException {

        Context subCtx = startCtx;
        StringTokenizer tokenizer = new StringTokenizer(subcontextPath, "/");
        int tokensNr = tokenizer.countTokens();
        String curToken = "";

        for ( int i=0; i < tokensNr; i++){
            curToken = tokenizer.nextToken();

            try {
                if ( subCtx.lookup(curToken) != null )
                    subCtx = (Context) subCtx.lookup(curToken);                    
            } catch (NameNotFoundException nnex){
                subCtx = subCtx.createSubcontext( curToken );
            }            
        }
        return subCtx;
    }


    /**
     * Sets up the env: context with the default values, like the java
     * system properties and os environment variable.
     */
    private static void initializeEnvContext(Hashtable properties) 
        throws NamingException {
        
        Enumeration en = properties.keys();
        String key;

        while ( en.hasMoreElements() ) {
            key = (String)en.nextElement();
            envBind( key, properties.get(key)); 
        }
    }

}
