/* 
 * $COPYRIGHT$
 * $Id: JabberClientConcept.java,v 1.4 2002/05/02 15:10:15 neuro Exp $
 *
 * Date        Author            Changes 
 * APR 08 2002 Szabo Csaba       Created
 */
package org.media.mn8.concepts;

import org.media.mn8.*;
import java.io.*;
import java.util.Stack;
import org.media.mn8.protocol.http.TagStriper;
import org.media.mn8.protocol.jabber.*;
import org.media.mn8.protocol.jabber.datablocks.*;

/**
 * @author <a href="mailto:crow@nolimits.ro">Szabo Csaba</a>
 * @version $Revision: 1.4 $ $Date: 2002/05/02 15:10:15 $
 */

public class JabberClientConcept extends JavaConcept implements JabberListener {

    //The resource name to present to the server
    private static final String JABBER_RESOURCE = "MN8-Client";
    
    //The re-establish connection number
    private static final int RECONNECT_NR = 5;
    
    //The number of seconds to waiting for the next connection
    private static final int WAIT_SEC = 10;

    //The stream to the jabber server
    private JabberStream jabberStream = null;
    
    //Flag to say whether or not we are connected to the server, and have been authenticated
    private boolean connectionOpen;

    //The name of the jabber server to connect to
    private String jabberServerName;

    //The port on which the jabber server runs on the machine named in "jabberServer"
    private int jabberServerPort;

    //The username to log into jabber server with
    private String jabberUserID;

    //The password to log into jabber server with
    private String jabberPasswd;


    private mn8ScriptMethod _method = null;
    private Concept        _concept = new NilConcept();

    //The responses element
    private TypedSeries       item = new TypedSeries( "item", "item" );  
    public ElementConcept response = new ElementConcept( "response", "response", item, null );
    


    /**
     * The default constructor
     */
    public JabberClientConcept() {
        //response.setValue( item );
    }


    /**
     * Method to set up the jabber server host and port
     *
     * @param host The name of the jabber server to connect to
     * @param port The port of the jabber server to connect to
     */ 
    public JabberClientConcept create( StringConcept host, IntegerConcept port ) {
        jabberServerName = host.getValue();
        jabberServerPort = port.getInt();
        return this;
    }
    
    
    /**
     * Method to log into Jabber server. 
     * This method established connection to jabber server.
     * 
     * @param user The username to log into jabber server with
     * @param pwd The password to log into jabber server with
     */
    public void login( StringConcept user, StringConcept pwd ) {
        jabberUserID = user.getValue();
        jabberPasswd = pwd.getValue();
        connectToJabberServer();
    }


    /**
     * Method to send to specified user the specified message
     *
     * @param to The user to send the message to
     * @param message The text message to send
     * @param html The html message to send 
     */
    public void send( StringConcept to, StringConcept message, StringConcept html ) {
        send( to.getValue().trim(), message.getValue().trim(), html.getValue().trim(), true );
    }

    
    /**
     * Method to send to specified user the specified message
     *
     * @param to The user to send the message to
     * @param message The message to send
     */
    public void send( StringConcept to, StringConcept message ) {
        String stripedMessage =  TagStriper.strip_tags( message.getValue().trim(), false );
        send( to.getValue().trim(), stripedMessage, 
              "<body>" + message.getValue().trim() + "</body>", 
              !stripedMessage.equals( message.getValue().trim() ) );
    }

    
    /**
     * The handler method
     *
     * @param concept The invoked method concept
     * @param methodName The handler method name (not a signature, it's will be formed here) 
     */
    public void addHandler( Concept concept, StringConcept methodName ) {
        _method  = (mn8ScriptMethod)concept.getConceptMethod( new StringConcept( methodName.toString() + ":JabberClient,Element" ) );
        _concept = concept;
    }

    
    /** 
     * Return the JabberConceptDefinition object
     */
    public ConceptDefinition getConceptDefinition() {
        return JabberClientDefinition.getReference();
    }
    
    
    /**
     * Clone this object
     */
    public Object clone() {
        return this;
    }


    /**
     * Method to connect to the jabber server
     */
    public synchronized void connectToJabberServer() {
        connectionOpen = false;
        if ( jabberStream != null ) jabberStream.close();
        jabberStream = null;

        try {
            ConnectorInterface connector = new seConnector( jabberServerName, jabberServerPort );
            jabberStream = new JabberStream( connector );
            jabberStream.setJabberListener( this );
        }
        catch ( Exception e ) {
            new ErrorConcept("error", "connectionFalied", e.getMessage() ); 
        }
    }

    
    /**
     * Method to being the conversation whit the
     */
    public void beginConversation() {
        try {
            Login login = new Login( jabberUserID, jabberPasswd, JABBER_RESOURCE );
            jabberStream.send( login );
        }        
        catch ( Exception e ) {
            new ErrorConcept("error", "connectionFalied", e.getMessage() ); 
        }
    }

    
    /**
     * Method to respond to the arrival of Jabber data block
     *
     * @param data The data block comming form the server
     */
    public void blockArrived( JabberDataBlock data ) {
        try {
            if ( data instanceof Iq ) {
                String result = (String)data.getAttribute( "type" );
                if ( result.equals( "result" ) ) {
                    sendPresence();
                    connectionOpen = true;
                }
            }
            if ( data instanceof Message ) {
                Message     message = (Message)data;
                SeriesConcept param = new SeriesConcept();
                ElementConcept   ec = new ElementConcept("item", "item", new StringConcept( message.getBody() ), null);
                
                ec.addAttribute( new AttributeConcept( "timestamp", "timestamp", 
                                                       ((DateDefinition)DateDefinition.getReference()).getToday() ) );

                ec.addAttribute( new AttributeConcept( "fromId", "from", new StringConcept( message.getAttribute("from"))));
                item.add( ec );
                if ( _method != null ) {
                    param.add(this );
                    param.add(ec );
                    _method.invoke( _concept, param );
                }
            }
        } 
        catch ( Exception e ) {
            new ErrorConcept("error", "connectionFailed", e.getMessage() );
        }
    }


    /**
     * Method to handle the termination of a connection
     *
     * @param e The exception that caused the connection to be terminated
     */
    public void connectionTerminated( Exception e ) {
        connectToJabberServer();
    }
    

    /**
     * Method to inform the jabber server that we are on-line
     */
    private void sendPresence() throws IOException {
        Presence pres = new Presence();
        jabberStream.send( pres );
    } 

    
    /**
     * Method to send normal or html text to specified user
     *
     * @param to The user to send the message to
     * @param text The text message to send
     * @param html The html message to send
     * @param isHTML is true if you can send a html message, false otherwise 
     */ 
    private void send( String to, String text, String html, boolean isHTML ) {
        if ( to.length() == 0 )
            new ErrorConcept( "warning", "recipientMissing", "No recipient has been specified" );
        if ( text.length() == 0 )
            new ErrorConcept( "warning", "messageMissing", "No message text has been specified" );
        
        int reconnectNr = RECONNECT_NR;
        while ( !connectionOpen && reconnectNr > 0 ) {
            try {
                Thread.sleep( WAIT_SEC*1000 );
            }
            catch ( InterruptedException e ){}
            if ( !connectionOpen ) connectToJabberServer();
            reconnectNr--;
        }

        if ( !connectionOpen ) 
            new ErrorConcept( "error", "connectionFailed", "Jabber Server unavailable" );

        try {
            Message outMessage = new Message( to, text );
            if ( isHTML ) outMessage.setHTMLBodyText( html );
            jabberStream.send( outMessage );
        }
        catch ( Exception e ) {
            if ( jabberStream != null ) {
                jabberStream.close();
                jabberStream = null;
            }
        }
    }
}
