/* 
 * $COPYRIGHT$
 * $Id: NamingConf.java,v 1.15 2002/06/04 17:10:38 atech Exp $
 *
 * Date        Author            Changes 
 * May 04 2001 Remus Pereni      Created
 */
package org.media.mn8.conf;


import javax.xml.parsers.SAXParserFactory;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.DocumentBuilder;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.xml.sax.SAXParseException;
import java.util.Hashtable;
import java.util.Vector;
import java.util.Properties;
import java.util.Enumeration;
import java.io.InputStream;
import java.io.File;
import org.media.mn8.LogManager;
import org.media.mn8.util.SimpleExpression;
import org.media.mn8.util.SystemEnvReader;
import org.media.mn8.util.StreamResolver;
import org.media.mn8.concepts.StringConcept;
import java.net.URL;

/**
 * Class who is responsible for loading and managing the 
 * configuration information stored in /conf/naning.mn8.xml
 * @author <a href="mailto:remus@nolimits.ro">Remus Pereni</a>
 * @version $Revision: 1.15 $ $Date: 2002/06/04 17:10:38 $
 */

public class NamingConf {
    /**
     * Where will be stored in our JNDI Namespace the output stream
     */
    public static final String _SYSTEM_OUT_URI = "env://system/out";

    /**
     * Where will be stored in our JNDI Namespace the error stream
     */
    public static final String _SYSTEM_ERR_URI = "env://system/err";

    /**
     * Where will be stored in our JNDI Namespace the input stream
     */
    public static final String _SYSTEM_IN_URI = "env://system/err";

    private static final String _NAMING_CONFIG_FILE = "/conf/mn8config.xml";
    private static NamingHandlerBase _namingHandler = new NamingHandlerBase();
    private static NamingConf _namingConf = new NamingConf();
    private static Hashtable _envLoad = null;

    
    /**
     * Dafault constructor
     */
    private NamingConf () {
        InputStream streamNaminXML = null;
        SAXParserFactory factory = SAXParserFactory.newInstance();
        URL resURL;

        try {
            Class _thisClass = Class.forName("org.media.mn8.conf.NamingConf");
            resURL = _thisClass.getResource( _NAMING_CONFIG_FILE );
            DocumentBuilderFactory domFactory = DocumentBuilderFactory.newInstance();
            domFactory.setValidating( false );
            DocumentBuilder domBuilder = domFactory.newDocumentBuilder();

            if ( resURL != null ) {
                factory.setValidating(true);
                SAXParser saxParser = factory.newSAXParser();
                saxParser.parse (resURL.toString(), _namingHandler);

                Document document = domBuilder.parse( resURL.toString());
                Node node = document.getDocumentElement().getFirstChild();
                while( node != null ) {
                    if( node.getNodeType() == Node.ELEMENT_NODE ) {
                        if( ((Element) node).getTagName().equalsIgnoreCase("concepts"))
                            ConceptsConf.getRef().initialize( (Element) node );
                        else if ( ((Element)node).getTagName().equalsIgnoreCase("java.properties") )
                            JavaPropertiesConf.getRef().initialize( (Element)node );
                        else if ( ((Element)node).getTagName().equalsIgnoreCase("context") ) {
                            ContextConf.getRef().initialize( ( Element)node );
                        }
                    }
                    node = node.getNextSibling();
                }
            } else
                throw new RuntimeException();
        } catch (ClassNotFoundException cex) {
            LogManager.message( "warning",
                                "conf.file_missing",
                                _NAMING_CONFIG_FILE );
           throw new RuntimeException();
        } catch (SAXParseException saxex) {
            LogManager.message( "warning",
                                "conf.parse_exception",
                                _NAMING_CONFIG_FILE +
                                "\n" + saxex.getMessage() );
            saxex.printStackTrace();
            throw new RuntimeException();        
        } catch (RuntimeException rex) {
            LogManager.message( "warning",
                                "conf.file_missing",
                                _NAMING_CONFIG_FILE );
           throw new RuntimeException();
        } catch (Exception ex) {
            ex.printStackTrace();
        }

    }


    public static void cleanAll() {
        _envLoad.clear();
    }

    
    /**
     * Method for getting the NamingConf instance. Don't forger that it
     * implements a Singleton pattern.
     * @return The default instance of NamingConf
     */
    public static NamingConf getReference () {
        return _namingConf;
    }


    /**
     * Returns an hashtable loaded with all the data contained in
     * the /conf/naming.mn8.xml file in the initial.context section.<br>
     * This hashtable should be enough to create an InitialContext.
     * @return An hashtable containing the properties needed to 
     * successfully create an JNDI InitialContext.
     */
    public Hashtable getInitialContextProperties () {
        return _namingHandler.getInitialContextProperties();
    }


    /**
     * Returns an hashtable containing all the java system variables as
     * the operating system environment variables, selected using the 
     * patterns from /conf/naming.mn8.xml the <load name="pattern*" /> and
     * <except name="patter*" /> elements.
     * @return A hastable containing the os and java environment variables.
     */    
    public Hashtable getEnvProperties() {
        Vector loadPatterns = _namingHandler.getEnvLoad(); 
        Vector exceptPatterns = _namingHandler.getEnvLoadExcept();
        Properties sysProp = System.getProperties();
        
        _envLoad = new Hashtable();

        if ( loadPatterns == null ) 
            loadPatterns = new Vector();
        
        if ( loadPatterns.size() < 1 )    // we load all the variables
            loadPatterns.add("*");

        if ( exceptPatterns == null ) 
            exceptPatterns = new Vector();
        
        _envLoad.put("env://system/out", new StringConcept( "System.out" ) );
        _envLoad.put("env://system/err", new StringConcept( "System.err") );
        _envLoad.put("env://system/in",  new StringConcept( "System.in") );
        
        // java system properties
        applyEnvLoadConditions (sysProp,
                                loadPatterns,
                                exceptPatterns);

        // operating system properties
        applyEnvLoadConditions(getSystemEnvironment(),
                               loadPatterns,
                               exceptPatterns);            
            
        return _envLoad;
    }


    /**
     * It is possible that mn8 will not use it's MemoryContext but rather it 
     * will use some external one. It is possible to map the memory context 
     * in any context desired in the external one providing a mapping which will
     * be applied.
     * @return The mapping into the external JNDI context of the env:// namespace.
     */
    public String getEnvMapping() {
        return _namingHandler.getEnvMapping();
    }


    /**
     * It is possible that mn8 will not use it's MemoryContext but rather it 
     * will use some external one. It is possible to map the memory context 
     * in any context desired in the external one providing a mapping which will
     * be applied.
     * @return The mapping into the external JNDI context of the mem:// namespace.
     */
    public String getMemMapping() {
        return _namingHandler.getMemMapping();
    }


    private void applyEnvLoadConditions(Properties from, 
                                        Vector loadPatterns, 
                                        Vector exceptPatterns) {
        Enumeration propEnum = from.propertyNames();
        String propKey;
        boolean isExcept;

        while ( propEnum.hasMoreElements() ) {
            propKey = (String) propEnum.nextElement(); 
            isExcept = false;
            
	    for ( int cnt = 0 ; cnt < loadPatterns.size() && !isExcept; cnt++ ) {
                if ( SimpleExpression.match ( propKey, (String)loadPatterns.elementAt(cnt) )) {
                    for (int i = 0; i < exceptPatterns.size() && !isExcept; i++) {
                        if ( SimpleExpression.match ( propKey, (String)exceptPatterns.elementAt(i) ) ) 
                            isExcept = true;
                    }
		}
	    }

	    if ( !isExcept ) _envLoad.put( "env://system/properties/" + propKey, 
					   new StringConcept( (String)from.getProperty(propKey) ) );
	}
    }

    
    private Properties getSystemEnvironment () {
        Hashtable osCmdEnvMap = _namingHandler.getEnvCommands();
        String currentOs = System.getProperty("os.name");
        String cmd="";
        Enumeration keys;
        Properties result = null;
        String tmp_str;

        if (  osCmdEnvMap == null ||
              osCmdEnvMap.isEmpty() ) return null;
       
        if ( osCmdEnvMap.containsKey(currentOs.toLowerCase()) ) {
            cmd = (String)osCmdEnvMap.get(currentOs.toLowerCase() );            
        }else { 
            /* we try to match it */
            keys = osCmdEnvMap.keys();
            while ( keys.hasMoreElements() ){
                tmp_str = (String) keys.nextElement();
                if ( currentOs.toLowerCase().indexOf( tmp_str ) != -1 )
                    cmd = (String)osCmdEnvMap.get( tmp_str );
            }
        }

        try {
            result = SystemEnvReader.getSystemEnvironment(cmd);
        } catch (Exception ex) {
            ex.printStackTrace();
        }    
        return result;
    }
}
