/* 
 * $COPYRIGHT$
 * $Id: JabberStream.java,v 1.1 2002/04/09 16:47:46 crow Exp $
 *
 * Date        Author            Changes 
 * APR 08 2002 Szabo Csaba       Created
 */
package org.media.mn8.protocol.jabber;

import java.io.*;
import java.util.*;
import org.media.mn8.protocol.jabber.datablocks.*;
import org.media.mn8.protocol.jabber.xmlparser.*;

/**
 * The stream to a jabber server.
 */
public class JabberStream implements XMLEventListener, Runnable {
    /**
     * The Output stream to the server.
     */
    private OutputStream outStream;
    
    /**
     * The input stream from the server.
     */
    private InputStream inpStream;
    
    /**
     * The dispatcher thread.
     */
    private JabberDataBlockDispatcher dispatcher;


    /**
     * Constructor. Connects to the server and sends the jabber welcome message.
     *
     * @param connectorInterface The connector which establishes the socket for
     * the connection
     */
    public JabberStream( ConnectorInterface connectorInterface ) throws IOException {
        this( connectorInterface, null, true );
    }


    /**
     * Constructor. Connects to the server and sends the jabber welcome message.
     *
     * @param connectorInterface The connector which establishes the socket for the connection
     * @param theListener The listener object to send all messages arriving on this stream to
     */
    public JabberStream( ConnectorInterface connectorInterface, JabberListener theListener ) throws IOException {
        this( connectorInterface, theListener, true );
    }


    /**
     * Constructor. Connects to the server and sends the jabber welcome message.
     *
     * @param connectorInterface The connector which establishes the socket for the connection
     * @param theListener The listener object to send all messages arriving on this stream to
     * @param connect Flag to say whether or not to connect during object instanciation (true = yes, false = no).
     */
    public JabberStream( ConnectorInterface connectorInterface, JabberListener theListener, boolean connect ) throws IOException {
        dispatcher = new JabberDataBlockDispatcher();
        if( theListener != null ) setJabberListener( theListener );
        if( connect == true ) open( connectorInterface );   
    }


    /**
     * Method to open the connection to the server.
     *
     * @param connectorInterface The connector which establishes the socket for the connection
     */
    public void open( ConnectorInterface connectorInterface ) throws IOException {
        outStream = connectorInterface.openOutputStream();
        inpStream = connectorInterface.openInputStream();

        StringBuffer streamInitiator = new StringBuffer( "<stream:stream to=\"" );
        String hostname = connectorInterface.getHostname();
        streamInitiator.append( hostname );
        streamInitiator.append( "\" xmlns=\"jabber:client\" xmlns:stream=\"http://etherx.jabber.org/streams\">" );
        outStream.write( streamInitiator.toString().getBytes() );
        outStream.flush();

        Thread newThread = new Thread( this );
        newThread.start();
    }


    /**
     * The threads run method. Handles the parsing of incomming data in its own thread.
     */
    public void run() {
        try {
            XMLParser parser = new XMLParser( this );
            DataInputStream dis = new DataInputStream( inpStream );
            parser.setInputUTF8Encoded( true );
            parser.parse( dis );
            dispatcher.broadcastTerminatedConnection( null );
        }
        catch( Exception e ) {
            e.printStackTrace();
            dispatcher.broadcastTerminatedConnection(e);
        }
    }


    /**
     * Method to close the connection to the server and tell the listener
     * that the connection has been terminated.
     */
    public void close() {
        dispatcher.setJabberListener( null );
        try {
            send( "</stream:stream>" );
            inpStream.close();
            outStream.close();
        }
        catch( IOException e ) {
            // Ignore an IO Exceptions because they mean that the stream is
            // unavailable, which is irrelevant.
        }
        finally {
            dispatcher.halt();
        }
    }

    
    /**
     * Method of sending data to the server.
     *
     * @param data The data to send.
     */
    public void send( byte[] data ) throws IOException {
        outStream.write( data );
        outStream.flush();
    }


    /**
     * Method of sending data to the server.
     *
     * @param The data to send to the server.
     */
    public void send( String data ) throws IOException {
        send( data.getBytes() );
    }


    /**
     * Method of sending a Jabber datablock to the server.
     *
     * @param block The data block to send to the server.
     */
    public void send( JabberDataBlock block ) throws IOException {
        send( block.getBytes() );
    }


    /**
     * Set the listener to this stream.
     */
    public void setJabberListener( JabberListener listener ) {
        dispatcher.setJabberListener( listener );
    }


    /**
     * The current class being constructed.
     */
    private JabberDataBlock currentBlock;

    /**
     * Method called when an XML tag is started in the stream comming from the server.
     *
     * @param name Tag name.
     * @param attributes The tags attributes.
     */
    public void tagStarted( String name, Hashtable attributes ) {
        if ( name.equals( "stream:stream" ) )
            dispatcher.broadcastBeginConversation();
        else if ( name.equals( "message" ) )
            currentBlock = new Message( currentBlock, attributes );
        else if ( name.equals("iq") )
            currentBlock = new Iq( currentBlock, attributes );
        else if ( name.equals("presence" ) )
            currentBlock = new Presence( currentBlock, attributes );
        else
            currentBlock = new JabberDataBlock( name, currentBlock, attributes );
    }


    /**
     * Method called when some plain text is encountered in the XML stream
     * comming from the server.
     *
     * @param text The plain text in question
     */
    public void plaintextEncountered( String text ) {
        if( currentBlock != null ) currentBlock.addText( text );   
    }


    /**
     * The method called when a tag is ended in the stream comming from the server.
     *
     * @param name The name of the tag that has just ended.
     */
    public void tagEnded( String name ) {
        if( currentBlock == null ) return;

        JabberDataBlock parent = currentBlock.getParent();
        if( parent == null )
            dispatcher.broadcastJabberDataBlock( currentBlock );
        else
            parent.addChild( currentBlock );
        currentBlock = parent;
    }
}
