/* 
 * $COPYRIGHT$
 * $Id: Message.java,v 1.1 2002/04/09 16:47:46 crow Exp $
 *
 * Date        Author            Changes 
 * APR 08 2002 Szabo Csaba       Created
 */
package org.media.mn8.protocol.jabber.datablocks;

import org.media.mn8.protocol.jabber.*;
import java.util.*;

/**
 * The class representing a Jabber message object
 */
public class Message extends JabberDataBlock {
 
    /**
     * Constructor. Prepares the message destination and body
     *
     * @param to The destination of the message
     * @param message The message text
     */
    public Message( String to, String message ) {
        super();
        if( to != null ) setAttribute( "to", to );
        if( message != null ) setBodyText( message );
    }


    /**
     * Constructor. Prepares the message destination
     *
     * @param to The destination of the message
     */
    public Message( String to ) {
        this( to, null );
    }


    /**
     * Default Constructor. Alls for construction of an empty message template.
     */
    public Message() {
        this( null );
    }


    /**
     * Constructor for incomming messages
     *
     * @param _parent The parent of this datablock
     * @param _attributes The list of element attributes
     */
    public Message( JabberDataBlock _parent, Hashtable _attributes ) {
        super( _parent, _attributes );
    }


    /**
     * Method to set the body text. Creates a block with body as it's tag name
     * and inserts the text into it.
     *
     * @param bodyText The string to go in the message body
     */
    public void setBodyText( String text ) {
        JabberDataBlock body = new JabberDataBlock( "body", null, null );
        body.addText( text );
        addChild( body );
    }


    /**
     * Method to set the body text written in HTML. Creates a block with html as
     * it's tag name in the xhtml name space and inserts the html into it.
     *
     * @param html The html to go in the message body
     */
    public void setHTMLBodyText( String html ) {
        JabberDataBlock body = new JabberDataBlock( "html", null, null );
        body.setAttribute( "xmlns", "http://www.w3.org/1999/xhtml" );
        body.addText( html );
        addChild( body );
    }


    /**
     * Method to set the message thread. Creates a block with thread as it's tag
     * name and inserts the thread name into it.
     *
     * @param threadName The string to go in the thread block
     */
    public void setThread( String text ) {
        JabberDataBlock thread = new JabberDataBlock( "thread", null, null );
        thread.addText( text );
        addChild( thread );
    }


    /**
     * Method to set the subject text. Creates a subject block and inserts the text into it.
     *
     * @param text The string to go in the message subject
     */
    public void setSubject( String text ) {
        JabberDataBlock subject = new JabberDataBlock( "subject", null, null );
        subject.addText( text );
        addChild( subject );
    }


    /**
     * Method to return the text for a given child block
     */
    private String getTextForChildBlock( String blockname ) {
        if( childBlocks == null ) return "";
        
        for( int i = 0 ; i < childBlocks.size() ; i++ ) {
            JabberDataBlock thisBlock = (JabberDataBlock) childBlocks.elementAt( i );
            if( thisBlock.getTagName().equals( blockname ) ) {
                return thisBlock.getText();
            }
        }

        return "";
    }


    /**
     * Method to get the message subject
     *
     * @return A string representing the message subject
   */    
    public String getSubject() {
        return getTextForChildBlock( "subject" );
    }


    /**
     * Method to get the message body
     *
     * @return The message body as a string
     */
    public String getBody() {
        return getTextForChildBlock( "body" );
    }


    /**
     * Construct a reply message
     *
     * @return A message object destined for the sender of this message with no subject or body
     */
    public Message constructReply() {
        if( attributes == null ) return null;

        Message reply = new Message( );
        super.constructReply( reply );

        String thread = getTextForChildBlock( "thread" );
        if( thread != null && thread.length() > 0 ) setThread( thread );

        return reply;
    }


    /**
     * Get the tag start marker
     *
     * @return The block start tag
     */
    public String getTagName() {
        return "message";
    }
}
