/* 
 * $COPYRIGHT$
 * $Id: CronTimeScheduler.java,v 1.3 2002/05/20 13:15:19 atech Exp $
 *
 * Date        Author            Changes 
 * May 16 2002 Antal Attila      Created
 */

package org.media.mn8.util;

import java.util.Hashtable;
import java.util.StringTokenizer;
import java.util.NoSuchElementException;
import org.media.mn8.Concept;
import org.media.mn8.mn8Loader;
import org.media.mn8.mn8ScriptMethod;
import org.media.mn8.concepts.ErrorConcept;
import org.media.mn8.concepts.StringConcept;
import org.media.mn8.concepts.SeriesConcept;


/**
 * CronTimeScheduler, waiting for the crons and runs the Entries.
 *
 * @author <a href="mailto:atech@nolimits.ro">Antal Attila</a>
 */
public class CronTimeScheduler implements Runnable {

    private boolean         m_running;
    private Hashtable       m_entries=new Hashtable() ;
    private SynchronizedHeap m_priorityQueue=new SynchronizedHeap( new BinaryHeap() );


    /**
     * Schedule a time based trigger.
     * Note that if a trigger already has same name then it is removed.
     *
     * @param name the name of the trigger
     * @param trigger the trigger
     * @param target the target
     */
    public synchronized void addTrigger( final String name,
										 final Cron trigger,
                                         final String target )
	throws Exception {
        try { removeTrigger( name ); }
        catch( final NoSuchElementException nse ) {}

        final CronScheduledEntry entry = new CronScheduledEntry( name, trigger, target );
		//	System.out.println("***: " + trigger.toString() );
        m_entries.put( name, entry );
        final boolean added = rescheduleEntry( entry, false );
        if( !added ) return;
        try {
            if( entry == m_priorityQueue.peek() ) {
                notifyAll();
            }
        }
        catch( final NoSuchElementException nse ) {
			throw new Exception("Unexpected exception: " + nse.getMessage());
        }
    }


    /**
     * Remove a scheduled trigger by name.
     *
     * @param name the name of the trigger
     * @exception NoSuchElementException if no trigger exists with that name
     */
    public synchronized void removeTrigger( String name )
        throws NoSuchElementException {
        final CronScheduledEntry entry = getEntry( name );
        entry.invalidate();
        m_entries.remove( name );
    }


    /**
     * Force a trigger time to be recalculated.
     *
     * @param name the name of the trigger
     * @exception NoSuchElementException if no trigger exists with that name
     */
    public synchronized void resetTrigger( final String name )
        throws NoSuchElementException {
        final CronScheduledEntry entry = getEntry( name );
        rescheduleEntry( entry, true );
    }


    /**
     * Reschedule an entry.
     * if clone is true then invalidate old version and create a new entry to
     * insert into queue.
     *
     * @param timeEntry the entry
     * @param clone true if new entry is to be created
     * @return true if added to queue, false if not added
     */
    private synchronized boolean rescheduleEntry( final CronScheduledEntry timeEntry,
                                                  final boolean clone ) {
        CronScheduledEntry entry = timeEntry;
        if( clone ) {
            entry = new CronScheduledEntry( timeEntry.getName(),
                                            timeEntry.getTimeTrigger(),
                                            timeEntry.getTarget() );
            timeEntry.invalidate();
            m_entries.remove( timeEntry.getName() );
            m_entries.put( timeEntry.getName(), entry );
        }

        final long next = entry.getTimeTrigger().getTimeAfter( System.currentTimeMillis() );

        if( 0 < next ) {
            entry.setNextTime( next );
            m_priorityQueue.insert( entry );
            if( entry == m_priorityQueue.peek() ) {
                notify();
            }
            return true;
        } else {
            return false;
        }
    }


    /**
     * Retrieve entry from set.
     *
     * @param name the name of entry
     * @return the entry
     * @exception NoSuchElementException if no entry is found with that name
     */
    public CronScheduledEntry getEntry( final String name )
        throws NoSuchElementException {
        final CronScheduledEntry entry = (CronScheduledEntry)m_entries.get( name );
        if( null != entry ) {
            return entry;
        } else {
            throw new NoSuchElementException();
        }
    }


    private void runEntry( final CronScheduledEntry entry ) {

        final Runnable runnable = new Runnable() {
                public void run() {
					String target = entry.getTarget();
					System.out.println("STARTING CRON: " + target );
					StringConcept cname = new StringConcept();
					SeriesConcept  param = new SeriesConcept();
					SeriesConcept iparam = new SeriesConcept();
					StringTokenizer st = new StringTokenizer(target);
					int idx=0;
					while (st.hasMoreTokens()) {
						String token = st.nextToken();
						if (idx==0) {
							cname = new StringConcept(token);
						} else {
							param.add(new StringConcept(token));
						}
						idx++;
					}

					Concept con  = mn8Loader.instantiateConcept( cname.getValue() );
					mn8ScriptMethod method = 
						(mn8ScriptMethod)con.getConceptMethod(new StringConcept("main:Series"));
					iparam.add(param);
					method.invoke(cname.getValue(), iparam);
                }
            };

        try {
			new Thread( runnable ).start();
		}
        catch( final Exception e ) {
            new ErrorConcept( "warning", "errExecTrigger",
							  "Error executing trigger " + entry.getName());
        }
    }


    public void start() throws Exception {
		new Thread( this ).start();
    }


    public void stop() {
        m_running = false;
        synchronized ( this ) {
            notifyAll();
        }
    }

    public void run() {
        m_running = true;

        while( m_running ) {
            long duration = 0;

            if( !m_priorityQueue.isEmpty() ) {
                CronScheduledEntry entry = null;
                synchronized( this ) {
                    entry = getNextEntry();
                    if( null == entry ) continue;

                    duration = entry.getNextTime() - System.currentTimeMillis();

                    if( duration < 0 ) {
                        m_priorityQueue.pop();
                    }
                }

                if( duration < 0 ) {
                    runEntry( entry );
                    
                    rescheduleEntry( entry, false );
                    continue;
                }                 
                else if( 0 == duration ) {
                    duration = 1;
                }
            }
            try {
				synchronized( this ) {
					Thread.sleep( duration );
				}
			}
            catch( final InterruptedException ie ) { }
        }
    }


    private synchronized CronScheduledEntry getNextEntry() {
        CronScheduledEntry entry = (CronScheduledEntry)m_priorityQueue.peek();
        while( !entry.isValid() ) {
            m_priorityQueue.pop();
            if ( m_priorityQueue.isEmpty() ) {
                return null;
            }
            entry = (CronScheduledEntry)m_priorityQueue.peek();
        }
        return entry;
    }

}

