/* 
 * $COPYRIGHT$
 * $Id: SimpleExpression.java,v 1.28 2002/03/22 18:38:24 atech Exp $
 *
 * Date        Author            Changes 
 * Apr 08 2001 Remus Pereni      Created
 */
package org.media.mn8.util;

import java.util.Vector;
import java.util.StringTokenizer;
import java.util.Collection;

/**
 * The SimpleExpression class implements a simple expression parser. 
 * The only operators of a simple expression are:
 * <ul>
 *   <li><b>*</b> - match everything, like the "." in regular 
 *       expressions</li>
 *   <li><b>#</b> - selects everything</li>
 * </ul>
 * It performs only two functions:
 * <ul>
 *    <li><b>select</b> -selects all the apparence of the strings marked 
 *         with the select ("#") operator</li>
 *    <li><b>match</b> - verifies if a string matches the pattern specified 
 *         using normal characters and the two operators.
 * </ul>
 * It is intended as a simple replacement for regular expression in the case 
 * of simple expressions.
 * @author <a href="mailto:remus@nolimits.ro">Remus Pereni</a>
 * @version $Revision: 1.28 $ $Date: 2002/03/22 18:38:24 $
 * @see String
 */
public class SimpleExpression {

    private static final String _MATCH_ALL  = "*";
    private static final String _SELECT_ALL = "#";

    /**
     * Selects a collection of strings from the target using as mask the pattern
     * and the two operators "#" and "*".
     * @param target The string to be matched
     * @param pattern The pattern
     * @return A collection (Vector) of string arrays containing the selected
     * strings as specified through the pattern or null of none find or 
     * specified.
     * @see java.util.Collection
     * @see java.util.Vector
     */
    public static Collection select (String target, String pattern) {
        Vector selection = null;
        int pattern_index = 0;
        int target_index = 0;
        String[] pattern_tokens = makeTokens (pattern); 
        String[] current_selection = new String [ operatorOccurences ( pattern, _SELECT_ALL) ];
        int la1_index;
        int select_op_index;        
        boolean failed = false;
        target = convert( target );
        
        if (!contains (target, pattern)) return selection;

    outer:
        while (target_index < target.length()) {

            for (pattern_index =0, select_op_index=0; 
                 pattern_index < pattern_tokens.length; 
                 pattern_index++) {                

                if ( pattern_tokens[pattern_index].equals( _MATCH_ALL ) ) {
                    
                    if ( pattern_index + 1  < pattern_tokens.length ) {
                        la1_index = lookAhead ( pattern_tokens[pattern_index + 1], 
                                                target, 
                                                target_index);
                        if ( la1_index != -1 ){
                            target_index = la1_index +  pattern_tokens[pattern_index + 1].length();
                            pattern_index ++;
                        }
                        else {
                            target_index = target.length();
                            /* there is another pattern token, yet not to be found in 
                               in target */
                            failed = true;
                            break;
                        }
                    }
                    else break;

                }                

                else  if ( pattern_tokens[pattern_index].equals( _SELECT_ALL ) ) {

                    if ( pattern_index + 1  < pattern_tokens.length ) {
                        la1_index = lookAhead ( pattern_tokens[pattern_index + 1], 
                                                target, 
                                                target_index);
                        if ( la1_index != -1 ) {
                            current_selection [select_op_index] = target.substring( target_index,
                                                                                    la1_index );
                            target_index = la1_index + pattern_tokens[pattern_index + 1].length();
                            select_op_index++;
                            pattern_index = pattern_index + 1;
                        }                    
                        else {
                            target_index = target.length();
                            /* there is another pattern token, yet not to be found in 
                               in target */
                            failed = true;
                            break;
                        }
                    }
                    else {
                        current_selection [select_op_index] = target.substring( target_index );
                        target_index = target.length();
                        break;
                    }
                } 

                else {
                    if ( target.startsWith(pattern_tokens[pattern_index], target_index) ) {
                        target_index += pattern_tokens[pattern_index].length();
                    }
                    else {
                        failed = true;
                        break outer;
                    }
                }         
            }

            if ( operatorOccurences ( pattern, _SELECT_ALL ) > 0 &&
                 !failed &&
                 current_selection[0] != null ) {                
                if ( selection== null ) selection = new Vector();
                selection.add( current_selection );
                current_selection = new String [ operatorOccurences ( pattern, _SELECT_ALL) ];     
            }
        }
        
        if ( target_index != target.length() ) {
            // not all the target characters where parsed, that means that either
            // the pattern is not vell specified or it does not match
            selection = null;
        }

        return selection;
    }


    /**
     * Specifies if a pattern matches the target string or not.
     * @param target The string to be matched
     * @param pattern The pattern
     * @return True if the pattern is present in the target, false otherway.
     */
    public static boolean match (String target, String pattern) {
        boolean matched = false;
        int pattern_index = 0;
        int target_index = 0;
        String[] pattern_tokens = makeTokens (pattern); 
        int crt_target_index =0;
        target = convert( target );

        if ( !isValidPattern (pattern) ) {
            throw new RuntimeException ( "Indetermination! There seems to be two consecutive " +
                                         "operators in your pattern (\"" + pattern + "\") , please revise it.");
        }
        
        // if simple expression, whithout operators looking for equality
        if ( pattern_tokens.length == 1 && !isOperator( pattern_tokens[0] ) )
            return target.equals( pattern );

        while ( !matched && target_index < target.length() ) {
            if ( isOperator(pattern_tokens[pattern_index]) ) pattern_index ++;
            else {
                crt_target_index = target.indexOf( pattern_tokens[pattern_index], target_index );
                if (crt_target_index == -1) return false;
                else {
                    target_index = crt_target_index + pattern_tokens[pattern_index].length();
                    pattern_index ++;
                }
            }
            if ( pattern_index >= pattern_tokens.length ) matched = true;
        }

        if ( ( isStartsWithMatch( target, pattern_tokens) || isStartsOpened( pattern_tokens, pattern_index) ) && 
             matched &&
             (target_index == target.length() || isOpenEnded(pattern_tokens, pattern_index) )  ) return true;
        else return false;
    }

    private static boolean isOpenEnded( String[] tokens, int index ) {
       if ( tokens[index-1].equals(_MATCH_ALL) ||  tokens[index-1].equals(_SELECT_ALL) ) return true;
       else return false;
    }


    private static boolean isStartsWithMatch( String target, String[] tokens ) {
       if ( target.startsWith( tokens[0] )  ) return true;
       else return false;
    }


    private static boolean isStartsOpened( String[] tokens, int index ) {
       if ( tokens[0].equals(_MATCH_ALL) ||  tokens[0].equals(_SELECT_ALL) ) return true;
       else return false;
    }


    private static String convert( String target ) {
        int cnt = 0;
        String result = "";
        if ( target.indexOf("\n\r") != -1 ) {
            StringTokenizer tokenizer = new StringTokenizer ( target, "\n\r" );
                 while ( tokenizer.hasMoreTokens() ) {
                    result += tokenizer.nextToken() +"\n";
                }
            return result;
       }
       else return target;
    }

   public static boolean contains (String target, String pattern) {
        boolean matched = false;
        int pattern_index = 0;
        int target_index = 0;
        String[] pattern_tokens = makeTokens (pattern); 
        int crt_target_index =0;
        target = convert( target );

        if ( !isValidPattern (pattern) ) {
            throw new RuntimeException ( "Indetermination! There seems to be two consecutive " +
                                         "operators in your pattern (\"" + pattern + "\") , please revise it.");
        }
        
        // if simple expression, whithout operators looking for equality
        if ( pattern_tokens.length == 1 && !isOperator( pattern_tokens[0] ) )
            return target.equals( pattern );

        while ( !matched && target_index < target.length() ) {
            if ( isOperator(pattern_tokens[pattern_index]) ) {
                pattern_index ++;
            }else {
                crt_target_index = target.indexOf( pattern_tokens[pattern_index], target_index );
                if (crt_target_index == -1) {
                    return false;
                } else {
                    target_index = crt_target_index + pattern_tokens[pattern_index].length();
                    pattern_index ++;
                }
            }

            if ( pattern_index >= pattern_tokens.length )
                matched = true;
        }
        return matched;
    }


    /**
     * Tries to find the token in the target string starting with the 
     * targetIndex position.
     * @param token The token to be looked for.
     * @param target The string where we look for the token.
     * @param targetIndex the position in the target where we begin the search.
     * @return The position of the first found token in the target or -1 in 
     *         case was not found.
     */
    protected static int lookAhead (String token, String target, int targetIndex) {
        return target.indexOf(token, targetIndex);
    }
    

    protected static String[] makeTokens (String pattern) {
        String[] result= null;
        int tokens = operatorOccurences (pattern, _MATCH_ALL) 
                    + operatorOccurences (pattern, _SELECT_ALL);

        if ( tokens == 0 ) {
            result = new String [1];
            result[0] = pattern;
        } else {           
            int cnt = 0;
            StringTokenizer tokenizer = new StringTokenizer ( pattern,
                                                              _MATCH_ALL + _SELECT_ALL,
                                                              true);
            result = new String [tokenizer.countTokens()];
            while ( tokenizer.hasMoreTokens() ) {
                result[cnt] = tokenizer.nextToken();
                cnt++;
            }
        }
        return result;
    }


    protected static int operatorOccurences (String target, String operator) {
        int currentIndex = 0;
        int occurence=0;
        
        while ( currentIndex != -1 ){
            currentIndex = target.indexOf (operator, currentIndex) ;
            if ( currentIndex != -1){
                occurence ++;
                currentIndex ++;
            }
        }
        return occurence;
    }


    protected static boolean isOperator (String target) {
        if ( target.equals ( _MATCH_ALL ) ||
             target.equals ( _SELECT_ALL ) )
            return true;
        else 
            return false;
    }


    private static boolean isValidPattern ( String pattern ) {
        if ( pattern.indexOf( _MATCH_ALL + _SELECT_ALL ) != -1 ||
             pattern.indexOf( _SELECT_ALL + _MATCH_ALL ) != -1 ||
             pattern.indexOf( _MATCH_ALL + _MATCH_ALL ) != -1  ||
             pattern.indexOf( _SELECT_ALL + _SELECT_ALL ) != -1 ) return false;
        else return true;
    }

}
