/* 
 * $COPYRIGHT$
 * $Id: SystemEnvReader.java,v 1.4 2002/02/20 18:01:38 neuro Exp $
 *
 * Date        Author            Changes 
 * May 15 2001 Remus Pereni      Created
 */
package org.media.mn8.util;


import java.util.Hashtable;
import java.util.Properties;
import java.util.Iterator;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.LineNumberReader;
import java.io.IOException;


/**
 * Simple class which executes an system command meant to list all
 * the system variables, parse the result of the command and returns
 * the system variables and their values in an map.
 * <p>
 * Invocation:<br>
 * <tt>SystemEnvReader.getSystemEnvironment()</tt><br>
 * In which case the "env" command is invoked. This is specific 
 * command for Unix environments.<br>
 * or <br>
 * <tt>SystemEnvReader.getSystemEnvironment(String cmd)</tt><br>
 * In which case the "cmd" command is invoked. This is the operating
 * system specific command to list all the system variables. For Windows 
 * this "cmd" could be: "cmd.exe /c set". <br><br>
 * <b>Note:</b> The parser expects that each variable is listed as it follows: <br>
 * <tt>VAR_NAME=VAR_VALUE EOL</tt>
 * </p>
 * @author <a href="mailto:remus@nolimits.ro">Remus Pereni</a>
 * @version $Revision: 1.4 $ $Date: 2002/02/20 18:01:38 $
 */
public class SystemEnvReader {

    private static final SystemEnvReader _envReader = new SystemEnvReader();


    /**
     * Returns the operating system specific environment variables and their values.
     * This method automatically tries to invoke the "env" command which is 
     * particular for Unix operating systems!.
     * @return A Properties containing the env variables name and value mappings.
     */
    public static Properties getSystemEnvironment()  throws IOException{        
        return _envReader.parseResult(_envReader.executeEnvCommand("env"));
    }


    /**
     * Returns the operating system specific environment variables and their values.
     * This method automatically tries to invoke the "cmd" command which is 
     * particular for the runtime operating system!
     * @param command The "env" or  "set" command particular to the operating 
     * system.
     * @return A Properties containing the env variables name and value mappings.
     */
    public static Properties getSystemEnvironment(String command)  throws IOException{        
        return _envReader.parseResult(_envReader.executeEnvCommand(command));
    }

    
    public static void main (String[] args) {
        try {
            Properties map;
            Iterator iterator;
            String key;

            if ( args.length > 0 )  
                map =  SystemEnvReader.getSystemEnvironment(args[0]);
            else
                map =  SystemEnvReader.getSystemEnvironment("env");

            if ( map.isEmpty() ) {
                System.out.println ("Usage: org.media.mn8.util.SystemEnvReader cmd \n" +
                                    "     where cmd is the env command on your op sys");
            }
                
            iterator = map.keySet().iterator();
            while ( iterator.hasNext() ) {
                key = (String)iterator.next();
                System.out.println(key + "=" + map.get(key));
            }

        }catch (Exception ex){
            ex.printStackTrace();
        }
    }


    private Properties parseResult( InputStream is) throws IOException {
        Properties result = new Properties();
        LineNumberReader lnreader = new LineNumberReader(new InputStreamReader(is));
        String line = "";
        String key;
        String value;

        while ( ( line = lnreader.readLine()) != null ) {
            if( line != null &&  line.indexOf("=") >= 0 ) {
                result.put( line.substring(0, line.indexOf("=")),
                            line.substring(line.indexOf("=") + 1));
            }
        }
        return result;
    }


    private InputStream executeEnvCommand (String command) throws IOException{
        Runtime r = Runtime.getRuntime();
        Process p = r.exec(command);
        return p.getInputStream();
    }


}
