/*
 * $COPYRIGHT$
 * $Id: ReferenceRefAddr.java,v 1.2 2001/06/04 10:23:21 neuro Exp $
 *
 * Date        Author            Changes 
 * May 20 2001 Remus Pereni      Created
 */

package org.media.naming;

import java.util.Hashtable;
import javax.naming.NamingException;
import javax.naming.RefAddr;
import javax.naming.Reference;
import javax.naming.Referenceable;
import javax.naming.spi.NamingManager;



/**
 * This class allows referenceable objects to be stored
 * inside of other references.
 * <p>
 * This class was inspired by the tyrex.naming package from the 
 * <a href="http://tyrex.exolab.org">Tyrex</a> project.
 * All credits for the good stuff should go to Riad Mohammed and
 * the <a href="http://www.exolab.org">Exolab</a> group, 
 * all the bad stuff blame it on 
 * <a href="mailto:remus@nolimits.ro">me</a> :).
 * </p>
 *
 * @author <a href="remus@nolimits.ro">Remus Pereni</a>
 * @version $Revision: 1.2 $ $Date: 2001/06/04 10:23:21 $
 */
final class ReferenceRefAddr 
    extends RefAddr
{


    /**
     * The reference
     */
    private final Reference _reference;


    /**
     * The environment used to recreate the referenceable
     * object from the underlying {@link #_reference}
     */
    private final Hashtable _environment;


    /**
     * The object
     */
    private transient Object _object;


    /**
     * The hash code
     */
    private transient int _hashCode;


    /**
     * Create the ReferenceRefAddr with the specified arguments.
     * <p>
     * The reference from the specified referenceable object is 
     * retrieved and stored.
     * <p>
     * A null environment hashtable is used in the recreation of the 
     * referenceable object from the underlying reference.
     *
     * @param addressType the address type of the ref addr
     * @param referenceable the referenceable object
     * @throw NamingException if there is a problem get the reference
     *      from the referenceable object.
     */
    ReferenceRefAddr (String addressType, Referenceable referenceable)
        throws NamingException
    {
        this(addressType, referenceable, null);
    }


    /**
     * Create the ReferenceRefAddr with the specified arguments.
     * <p>
     * The reference from the specified referenceable object is 
     * retrieved and stored.
     *
     * @param addressType the address type of the ref addr
     * @param referenceable the referenceable object
     * @param environment used in the recreation of the
     * referenceable object from the underlying reference. Can be null.
     * @throw NamingException if there is a problem get the reference
     *      from the referenceable object.
     */
    ReferenceRefAddr (String addressType, 
                      Referenceable referenceable,
                      Hashtable environment)
        throws NamingException
    {
        super(addressType);

        if (null == referenceable) {
            throw new IllegalArgumentException("The argument 'referenceable' is null.");
        }

        _reference = referenceable.getReference();
        _environment = environment;

    }


    /**
     * Create the ReferenceRefAddr with the specified arguments.
     * <p>
     * A null environment hashtable is used in the recreation of the 
     * referenceable object from the underlying reference.
     *
     * @param addressType the address type of the ref addr
     * @param reference the reference
     */
    ReferenceRefAddr( String addressType, Reference reference )
    {
        this(addressType, reference, null);
    }

    /**
     * Create the ReferenceRefAddr with the specified arguments.
     *
     * @param addressType the address type of the ref addr
     * @param reference the reference
     * @param environment used in the recreation of the
     * referenceable object from the underlying reference. Can be null.
     */
    ReferenceRefAddr( String addressType, 
                      Reference reference,
                      Hashtable environment )
    {
        super( addressType );

        if (null == reference)
            throw new IllegalArgumentException("The argument 'reference' is null.");
        _reference = reference;
        _environment = environment;

    }

    /**
     * Return true if the specified object is of type
     * ReferenceRefAddr and has the same underlying reference.
     *
     * @param object the object
     * @return true if the specified object is of type
     * ReferenceRefAddr and has the same underlying reference.
     */
    public boolean equals(Object object)
    {
        if ((null == object) ||
            !(object instanceof ReferenceRefAddr)) {
            return false;
        }

        return _reference.equals(((ReferenceRefAddr)object)._reference);
    }

    /**
     * Return the hashcode of the underlying reference as the hashcode
     * of the ReferenceRefAddr.
     *
     * @return the hashcode of the underlying reference as the hashcode
     *      of the ReferenceRefAddr.
     */
    public int hashCode()
    {
        if (0 == _hashCode) 
            _hashCode = _reference.hashCode();

        return _hashCode;
    }

    /**
     * Returns the object referred to by the underlying
     * reference.
     *
     * @return the object referred to by the underlying
     * reference.
     */
    public Object getContent()
    {
        if (null == _object) {
            try {
                _object = NamingManager.getObjectInstance(_reference, null, null, _environment);        
            } catch (Exception e) { }
        }

        return _object;
    }

/*
    public static class Test 
        implements Referenceable, javax.naming.spi.ObjectFactory
    {
        private final String _string;

        public Test()
        {
            this("");
        }

        private Test(String string)
        {
            _string = string;
        }

        public javax.naming.Reference getReference() 
        {
            Reference ref = new Reference("org.media.naming.ReferenceRefAddr$Test", 
                                          "org.media.naming.ReferenceRefAddr$Test", 
                                          null);
            ref.add(new javax.naming.StringRefAddr("test", _string));
            return ref;
        }

        public Object getObjectInstance(Object object, 
                                                  javax.naming.Name name, 
                                                  javax.naming.Context context, 
                                                  Hashtable env)
            throws Exception
        {
            Reference ref = (Reference)object;
            return new Test((String)ref.get("test").getContent());
        }

        public String toString()
        {
            return super.toString() + "[" + _string + "]";
        }
    }


    public static class TestRef
        implements Referenceable, javax.naming.spi.ObjectFactory
    {
        private final Test _test;

        public TestRef()
        {
            this(null);
        }

        private TestRef(Test test)
        {
            _test = test;
        }

        public javax.naming.Reference getReference() 
        {
            Reference ref = new Reference("org.media.naming.ReferenceRefAddr$TestRef", 
                                          "org.media.naming.ReferenceRefAddr$TestRef", 
                                          null);
            try {
                ref.add(new ReferenceRefAddr("testref", _test));
            }
            catch(NamingException e) {
                e.printStackTrace();
            }
            return ref;
        }

        public Object getObjectInstance(Object object, 
                                                  javax.naming.Name name, 
                                                  javax.naming.Context context, 
                                                  Hashtable env)
            throws Exception
        {
            Reference ref = (Reference)object;
            return new TestRef((Test)ref.get("testref").getContent());
        }

        public String toString()
        {
            return super.toString() + "[" + _test + "]";
        }
    }

    public static void main (String args[]) 
        throws Exception
    {
        javax.naming.InitialContext ctx = new javax.naming.InitialContext();
        TestRef testref = new TestRef(new Test("Riad was here"));
        System.out.println("before " + testref);

        ctx.bind("testref", testref);

        TestRef resolved = (TestRef)ctx.lookup("testref");

        System.out.println("after " + resolved);
        
    }
*/
}
